#pragma once
#ifndef LRTABLES_LR0ITEMSETBASE_H_GGIPISTD
#define LRTABLES_LR0ITEMSETBASE_H_GGIPISTD

#include "Parsodus/lrtables/LR0Item.h"

namespace pds {
namespace lr {

/**
 * Provide common methods for Itemsets based upon LR(0) items
 * (SLR(1) and LR(0))
 */
template <typename Itemset>
class LR0ItemsetBase {
public:
    LR0ItemsetBase();
    LR0ItemsetBase(std::shared_ptr<Rule> start);

    void close(const Grammar& g);
    Itemset succ(std::string sym) const;
    bool operator==(const Itemset& rhs) const;
    bool canMerge(const Itemset& rhs) const;
    void merge(const Itemset& rhs);
    bool empty() const;

protected:
    std::set<LR0Item> m_items;
};

template <typename Itemset>
LR0ItemsetBase<Itemset>::LR0ItemsetBase()
{}

template <typename Itemset>
LR0ItemsetBase<Itemset>::LR0ItemsetBase(std::shared_ptr<Rule> start) {
    m_items.emplace(LR0Item{start, 0});
}

template <typename Itemset>
void LR0ItemsetBase<Itemset>::close(const Grammar& g) {
    bool changes = true;
    std::vector<LR0Item> todo;
    std::set<std::string> added;

    while (changes) {
        changes = false;
        std::set<LR0Item> toAdd;

        for (const LR0Item& i : m_items) {
            if (i.dotIdx < i.rule->tail.size()) {
                std::string& sym = i.rule->tail[i.dotIdx];
                if (g.variables.count(sym) && !added.count(sym)) {
                    added.insert(sym);
                    changes = true;
                    for (const auto& rule : g.rules) {
                        if (rule->head == sym) {
                            toAdd.insert(LR0Item{rule, 0});
                        }
                    }
                }
            }
        }
        m_items.insert(toAdd.begin(), toAdd.end());
    }
}

template <typename Itemset>
Itemset LR0ItemsetBase<Itemset>::succ(std::string sym) const {
    Itemset sc;
    for (auto& item : m_items) {
        if (item.dotIdx < item.rule->tail.size()) {
            if (item.rule->tail[item.dotIdx] == sym) {
                sc.m_items.insert(LR0Item{item.rule, item.dotIdx + 1});
            }
        }
    }
    return sc;
}

template <typename Itemset>
bool LR0ItemsetBase<Itemset>::operator==(const Itemset& rhs) const {
    return m_items == rhs.m_items;
}

template <typename Itemset>
bool LR0ItemsetBase<Itemset>::canMerge(const Itemset&) const {
    return false;
}

template <typename Itemset>
void LR0ItemsetBase<Itemset>::merge(const Itemset&) {
    //NO-OP
}

template <typename Itemset>
bool LR0ItemsetBase<Itemset>::empty() const {
    return m_items.empty();
}
    

} /* lr */ 
} /* pds */

#endif /* LRTABLES_LR0ITEMSETBASE_H_GGIPISTD */
